from gameoflife.game.Cell import Cell
from gameoflife.game.Form import Form
import random


class Runner:
    '''Class wich provide a map of cells and
     manage their behaviour through rules'''

    to_revive = []
    to_kill = []
    forms = Form

    def __init__(self, forms_amount=50, density=100, size=50, initial_status="dead", rounds=50):
        self.density = density
        self.size = size
        self.map = self.generate_map(initial_status, size)
        self.set_addresses()
        self.generate_forms(forms_amount)
        self.rounds = rounds

    def generate_map(self, initial_status, size):
        return [[Cell((i + 1) * (_ + 1), initial_status) for i in range(size)] for _ in range(size)]

    def generate_forms(self, amount):
        for _ in range(amount):
            tries = 10
            shape = Form.get_shape()
            placable = False
            while placable is not True and tries > 0:
                tries -= 1
                random_point = self.get_random_point()
                placable = self.check_place(random_point, shape)
                if placable:
                    self.place(random_point, shape)

        return

    def check_place(self, origin, shape):
        for x_index in range(Form.get_shape_form(shape)["x"]):
            for y_index in range(Form.get_shape_form(shape)["y"]):
                x = (origin['x'] + x_index) % self.size
                y = (origin['y'] + y_index) % self.size
                if self.is_free(x, y) is not True:
                    return False
        return True

    def place(self, origin, shape):
        for x_index in range(Form.get_shape_form(shape)["x"]):
            for y_index in range(Form.get_shape_form(shape)["y"]):
                x = (origin['x'] + x_index) % self.size
                y = (origin['y'] + y_index) % self.size
                self.map[y][x].status = 'alive' if shape[y_index][x_index] == 'a' else 'dead'

        return

    def is_free(self, x, y):
        x %= self.size
        y %= self.size
        return True if self.map[y][x].status == 'dead' else False

    def get_random_point(self):
        return {
            'x': random.randint(0, self.size - 1),
            'y': random.randint(0, self.size - 1)
        }

    def set_addresses(self):
        for i in range(len(self.map)):
            y = i
            for j in range(len(self.map[i])):
                cell = self.map[i][j]
                x = j
                cell.set_address(x, y)

    def run(self):
        if self.rounds < 0:
            return

        for cell_line in self.map:
            for cell in cell_line:
                self.try_kill(cell) if cell.status == 'alive' else self.try_revive(cell)
        self.update()
        self.reset()
        self.rounds -= 1

    def try_kill(self, cell):
        neighbourhood = self.get_neighborhood(cell)
        if len(neighbourhood) != 2 and len(neighbourhood) != 3:
            self.to_kill.append(cell)

    def try_revive(self, cell):
        neighbourhood = self.get_neighborhood(cell)
        if len(neighbourhood) == 3:
            self.to_revive.append(cell)

    def get_neighborhood(self, cell):
        neigbhors = []
        top_y_index = cell.address['y'] - \
            1 if cell.address['y'] > 0 else self.size - 1
        bot_y_index = cell.address['y'] + \
            1 if cell.address['y'] < self.size - 1 else 0

        for y in [top_y_index, cell.address['y'], bot_y_index]:
            left_x_index = cell.address['x'] - \
                1 if cell.address['x'] > 0 else self.size - 1
            right_x_index = cell.address['x'] + \
                1 if cell.address['x'] < self.size - 1 else 0
            for x in [left_x_index, cell.address['x'], right_x_index]:
                neigbhors.append(self.map[y][x])

        return [n for n in neigbhors if n.status == 'alive' and n is not cell]

    def update(self):
        for cell in self.to_revive:
            cell.revive()

        for cell in self.to_kill:
            cell.die()

    def reset(self):
        self.to_kill = []
        self.to_revive = []

    def __str__(self):
        return ('\ndensity = ' + str(self.density) + '\nsize = ' + str(self.size) + '\npopulation = \n' + self.stringify_map())

    def stringify_map(self):
        full_cells = [[str(cell) for cell in cell_array]
                      for cell_array in self.map]
        for i in range(len(full_cells)):
            print(full_cells[i])

        return str(full_cells)



import pygame

class GameOfLife:
    '''Class used to setup a pygame window and
    display all of the runner map states'''

    window_surface = None

    def __init__(self, runner):
        pygame.init()
        self._runner = runner
        self._death_color = (0, 0, 0)
        self._alive_color = (255, 0, 0)
        self._scale = 8

    def launch(self):
        self.window_surface = pygame.display.set_mode((self._runner.size*self._scale,self._runner.size*self._scale))

        launched = True
        while launched:
            self.treatment(self._runner)
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    launched = False

    def treatment(self, runner):
        for elements in runner.map:
            for element in elements:
                color = self._alive_color if element.status == 'alive' else self._death_color
                cell = pygame.Rect(element.address['x'] * self._scale, element.address['y'] * self._scale, self._scale, self._scale)
                pygame.draw.rect(self.window_surface, color, cell)

            pygame.display.flip()
        runner.run()


class Cell:
    '''Class used to represent a cell with a state and it address on a map'''

    def __init__(self, n, status='dead'):
        self.status = status
        self.n = n
        self.address = {'x': 'unset', 'y': 'unset'}

    def __str__(self):
        return 'X' if self.status == 'dead' else 'V'

    def revive(self):
        self.status = 'alive'

    def die(self):
        self.status = 'dead'

    def set_address(self, x, y):
        self.address['x'] = x
        self.address['y'] = y



from tkinter import StringVar, Label, Entry, Button
from functools import partial

class IHM:
    '''Class used to generate an Ihm
    which give parameter to a runner instance through user inputs'''

    def __init__(self):
        self._values = list()
        self._replay = True

    def _get_values(self):
        for c, v in enumerate(self._values):
            try:
                self._values[c] = int(v)
            except ValueError:
                print('No letter allowed', v)
                return False
            except TypeError:
                print('No letter allowed', v)
                return False
        return self._values

    def _get_replay(self):
        return self._replay

    def _set_replay(self, bool):
        self._replay = bool

    values = property(_get_values)
    replay = property(_get_replay, _set_replay)

    def data_treatment(self, elements, values, root):
        values.clear()
        for i in elements:
            values.append(i[1].get())
        root.destroy()

    def close(self, root):
        self._set_replay(False)
        root.destroy()

    def retry(self, root):
        root.destroy()

    def launch(self, root):
        root.title("Paramètres GOL")
        root.geometry("300x150")
        title = Label(root, text='PARAMETERS :')
        elements = [(Label(root, text='Dimension'), Entry(root, textvariable=(StringVar(root)))),
                    (Label(root, text='Round number'), Entry(root, textvariable=(StringVar(root)))),
                    (Label(root, text='Form amount'), Entry(root, textvariable=(StringVar(root))))]
        button = Button(root, text='lancer', command=partial(self.data_treatment, elements, self._values, root))

        title.grid(column=0, row=0)

        [(value[1][0].grid(column=0, row=value[0]+1), value[1][1].grid(column=1, row=value[0]+1)) for value in [(i,tupl) for i,tupl in enumerate(elements)]]

        button.grid(column=1, row=5)
        root.mainloop()

    def pop_up(self, root):
        root.title("Rejouer au GOL")
        root.geometry("250x100")
        title_retry = Label(root, text='Would you like to retry ?')
        buttonYes = Button(root, text='Yes', command=partial(self.retry, root))
        buttonNo = Button(root, text='No', command=partial(self.close, root))

        title_retry.grid(column=0, row=0)

        buttonYes.grid(column=0, row=1)
        buttonNo.grid(column=1, row=1)
        root.mainloop()



from enum import Enum
import random

class Form(Enum):
    '''Class containing every cells shape'''

    STABLE = {
        'BLOC': [
            ['a', 'a'],
            ['a', 'a']
        ],
        'PIPE': [
            ['d', 'a', 'd'],
            ['a', 'd', 'a'],
            ['d', 'a', 'd']
        ],
        'BOAT': [
            ['a', 'a', 'd'],
            ['a', 'd', 'a'],
            ['d', 'a', 'd']
        ],
        'SNAKE': [
            ['a', 'a', 'd', 'a'],
            ['a', 'd', 'a', 'a']
        ],
    }

    OSCILLATOR = {
        'BLINKER1': [
            ['a', 'a', 'a']
        ],
        'BLINKER2': [
            ['a'],
            ['a'],
            ['a']
        ]
    }

    SHIP = {
        'ANTS': [
            ['a', 'd', 'a'],
            ['d', 'd', 'a'],
            ['a', 'a', 'a']
        ]
    }

    @classmethod
    def get_shape_form(cls, shape):
        return {
            'x': len(shape[0]),
            'y': len(shape)
        }

    @classmethod
    def get_shape(cls, form_type=None):
        if(form_type is None):
            form_type = random.choice(list(Form))
        random_shape = random.choice(list(form_type.value))
        return form_type.value[random_shape]


import os

import setuptools
from setuptools import setup

setup(
    name='Valle_gameoflife',
    version='1.0.5',
    description='Conway\'s Game of Life for ESGI Python project.',
    url='https://github.com/LeDixiemeHomme/Python-Game-Of-Life',
    author='2ndGroupe',
    packages=setuptools.find_packages(),
    scripts=[
        'bin/gameoflife/',
        'bin/gameoflife.bat',
    ],
    zip_safe=False,
    install_requires=[
        'pygame'
    ]
)
